/*********************************************************************NVMH3****

Copyright NVIDIA Corporation 2002
TO THE MAXIMUM EXTENT PERMITTED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED
*AS IS* AND NVIDIA AND ITS SUPPLIERS DISCLAIM ALL WARRANTIES, EITHER EXPRESS
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL NVIDIA OR ITS SUPPLIERS
BE LIABLE FOR ANY SPECIAL, INCIDENTAL, INDIRECT, OR CONSEQUENTIAL DAMAGES
WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
ARISING OUT OF THE USE OF OR INABILITY TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.


Comments:

******************************************************************************/

//
// gcd.cpp
//
// basic example of the use of the Cg runtime in a simple OpenGL program
//

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <OpenGL/glext.h>
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>
#include <GL/glut.h>
#endif
#include <Cg/cg.h>
#include <Cg/cgGL.h>

#define RES 512

// globals holding the Cg context that we're storing our programs in as well
// as handles to the vertex and fragment programs used
CGcontext	context;
CGprogram	vertexProgram, fragmentProgram;
CGprofile	vertexProfile, fragmentProfile;

// globals
int		lod=4;
double		dx=0.0,dy=0.0;

// global texture ids
const int	textures=2;
unsigned long	texID[] = {GL_TEXTURE0_ARB,GL_TEXTURE1_ARB};
GLuint		texNames[2];

// forward declarations of functions
int	main(int argc, char *argv[]);
void	interpmenuhandler(int c);
void	lodmenuhandler(int l);
void	mainmenuhandler(int m);
void	mousemove(int x, int y);
void	keyboard(unsigned char key, int x, int y);
void	display();
void	handleCgError();
void	initializeGL();
void	chooseCgProfiles();
void	loadCgPrograms();
void	imageTexture();
void	maskTexture();
float	gauss(int i, int j, float sigma);
float	width();
float	height();

int main(int argc, char *argv[])
{
	int	mainmenu,interpmenu,lodmenu;
	
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
  glutInitWindowSize(RES, RES);
  glutCreateWindow("Cg gcd");

  initializeGL();

  interpmenu = glutCreateMenu(interpmenuhandler);
  glutAddMenuEntry("None",1);
  glutAddMenuEntry("Intra-map",2);
  glutAddMenuEntry("Inter-map",3);
  glutAddMenuEntry("Both",4);

  lodmenu = glutCreateMenu(lodmenuhandler);
  glutAddMenuEntry("1",1);
  glutAddMenuEntry("2",2);
  glutAddMenuEntry("3",3);
  glutAddMenuEntry("4",4);

  mainmenu = glutCreateMenu(mainmenuhandler);
  glutAddSubMenu("Interpolation",interpmenu);
  glutAddSubMenu("LOD",lodmenu);
  glutAddMenuEntry("Quit",9);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  glutMotionFunc(mousemove);
  glutKeyboardFunc(keyboard);
  glutDisplayFunc(display);

  // and all the rest happens in the display function...
  glutMainLoop();

  return 0;
}

void mainmenuhandler(int m)
{
  switch(m) {
    case 9:
      cgDestroyProgram(vertexProgram);
      cgDestroyProgram(fragmentProgram);
      cgDestroyContext(context);
      exit(0);
    break;
  }
}

void interpmenuhandler(int c)
{
//fprintf(stderr,"interp: %d\n",c);
  //                                                       (Manhattan dist.)
  //                                   intra-map interp    inter-map interp
  // GL_TEXTURE_MIN_FILTER options:
  //   GL_NEAREST_MIPMAP_NEAREST:             no                  no
  //   GL_LINEAR_MIPMAP_NEAREST              yes                  no
  //   GL_NEAREST_MIPMAP_LINEAR               no                 yes
  //   GL_LINEAR_MIPMAP_LINEAR               yes                 yes
  // GL_TEXTURE_MAG_FILTER options:
  //   GL_NEAREST                             no                 n/a
  //   GL_LINEAR                             yes                 n/a
  glActiveTextureARB(texID[0]);
  switch(c) {
    case 1:
      // no smoothing
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_NEAREST_MIPMAP_NEAREST);
      break;
    case 2:
      // intra-map
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_LINEAR_MIPMAP_NEAREST);
      break;
    case 3:
      // inter-map
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_NEAREST_MIPMAP_LINEAR);
      break;
    case 4:
    default:
      // both
      glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,
                                    GL_LINEAR_MIPMAP_LINEAR);
      break;
  }

  glutPostRedisplay();
}

void lodmenuhandler(int l)
{
  lod = l;
  glutPostRedisplay();
}

void mousemove(int x, int y)
{ 
//fprintf(stderr,"mousemove: (%d,%d)\n",x,y);
  dx = width()/2.0 - (float)x;
  dy = (float)y - height()/2.0;
  dx /= width();
  dy /= height();
//fprintf(stderr,"mousemove: (%4.2f,%4.2f)\n",(float)dx,(float)dy);
  glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
  switch(key) {
    case 'q':
    case 'Q':
    case 27:
      cgDestroyProgram(vertexProgram);
      cgDestroyProgram(fragmentProgram);
      cgDestroyContext(context);
      exit(0);
      break;
    case '1':
      lod = 1;
      break;
    case '2':
      lod = 2;
      break;
    case '3':
      lod = 3;
      break;
    case '4':
      lod = 4;
      break;
  }
  glutPostRedisplay();
}

void handleCgError() 
{
  fprintf(stderr, "Cg error: %s\n", cgGetErrorString(cgGetError()));
//exit(1);
}

void initializeGL()
{
  fprintf(stderr,"GL: %s\n",glGetString(GL_VERSION));

  // basic Cg setup; register a callback function for any errors
  // and create an initial context
  cgSetErrorCallback(handleCgError);
  context = cgCreateContext();
  chooseCgProfiles();
  loadCgPrograms();

  glGenTextures(textures,texNames);
  imageTexture();
  maskTexture();

  glClearColor(.25, .25, .25, 1.);
}

void display() 
{
	float	lx,ly;
	float	lw,lh;

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0,width(),0.0,height());

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  glDrawBuffer(GL_BACK);
  glClear(GL_COLOR_BUFFER_BIT);

  // make sure that the vertex and fragment programs, loaded
  // in loadCgPrograms() are bound
  cgGLEnableProfile(vertexProfile);
  cgGLEnableProfile(fragmentProfile);
  cgGLBindProgram(vertexProgram);
  cgGLBindProgram(fragmentProgram);

  // bind uniform parameters to vertex shader
  cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram,
                              "ModelViewProj"),
                              CG_GL_MODELVIEW_PROJECTION_MATRIX,
                              CG_GL_MATRIX_IDENTITY);
  cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram,
                              "ModelView"),
                              CG_GL_MODELVIEW_MATRIX,
                              CG_GL_MATRIX_IDENTITY);
  cgGLSetStateMatrixParameter(cgGetNamedParameter(vertexProgram,
                              "ModelViewIT"),
                              CG_GL_MODELVIEW_MATRIX,
                              CG_GL_MATRIX_INVERSE_TRANSPOSE);

  cgGLSetParameter1f(cgGetNamedParameter(fragmentProgram,"min_lod"),(float)lod);

  // texID[0]: image
  cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram,"img_tex"));

  // texID[1]: mask
  cgGLEnableTextureParameter(cgGetNamedParameter(fragmentProgram,"deg_tex"));

  lx = (float)dx;
  ly = (float)dy;
//fprintf(stderr,"(lx,ly): (%4.2f,%4.2f)\n",lx,ly);
  // not sure how Nathan's 'aspect ratio' calculation works here yet...
  // (1,1) gives the proper offset for square windows and square masks
  // matching the window size...
//lw = (float)width() / mask->cols();
//lh = (float)height() / mask->rows();
  lw = 1.0;
  lh = 1.0;
//fprintf(stderr,"(lw,lh): (%4.0f,%4.0f)\n",lw,lh);

  // render
  glBegin(GL_QUADS);
  {
    glMultiTexCoord2fARB(texID[0],0,0);
    glMultiTexCoord2fARB(texID[1],lx,ly);
    glVertex2f(0,0);

    glMultiTexCoord2fARB(texID[0],1,0);
    glMultiTexCoord2fARB(texID[1],lx+lw,ly);
    glVertex2f(width(),0);

    glMultiTexCoord2fARB(texID[0],1,1);
    glMultiTexCoord2fARB(texID[1],lx+lw,ly+lh);
    glVertex2f(width(),height());

    glMultiTexCoord2fARB(texID[0],0,1);
    glMultiTexCoord2fARB(texID[1],lx,ly+lh);
    glVertex2f(0,height());
  }
  glEnd();

  cgGLDisableTextureParameter(cgGetNamedParameter(fragmentProgram,"deg_tex"));

  cgGLDisableTextureParameter(cgGetNamedParameter(fragmentProgram,"img_tex"));

  cgGLDisableProfile(fragmentProfile);
  cgGLDisableProfile(vertexProfile);

  glutSwapBuffers();
}

void chooseCgProfiles()
{
  // choose the latest vertex and fragment profiles to use
//vertexProfile = cgGLGetLatestProfile(CG_GL_VERTEX);
//fragmentProfile = cgGLGetLatestProfile(CG_GL_FRAGMENT);
//cgGLSetOptimalOptions(vertexProfile);
//cgGLSetOptimalOptions(fragmentProfile);
//fprintf(stderr, "Cg vertex profile: ");
//switch(vertexProfile) {
//  case CG_PROFILE_ARBVP1:	fprintf(stderr, "arbvp1.\n"); break;
//  case CG_PROFILE_VP30:	fprintf(stderr, "vp30.\n"); break;
//  case CG_PROFILE_VP20:	fprintf(stderr, "vp20.\n"); break;
//  default:			fprintf(stderr, "unknown.\n"); break;
//}
//fprintf(stderr, "Cg fragment profile: ");
//switch(fragmentProfile) {
//  case CG_PROFILE_ARBFP1:	fprintf(stderr, "arbfp1.\n"); break;
//  case CG_PROFILE_FP30:	fprintf(stderr, "fp30.\n"); break;
//  case CG_PROFILE_FP20:	fprintf(stderr, "fp20.\n"); break;
//  default:			fprintf(stderr, "unknown.\n"); break;
//}

//return;

  // choose the vertex and fragment profiles to use.  try to use
  // CG_PROFILE_ARBVFP1 and CG_PROFILE_ARBFP1, depending on hardware support.
  // if those aren't available, fall back to CG_PROFILE_VP30 and
  // CG_PROFILE_FP30, respectively.

  // make sure that the appropriate profiles are available on the user's system.
  if (cgGLIsProfileSupported(CG_PROFILE_ARBVP1)) {
    vertexProfile = CG_PROFILE_ARBVP1;
    fprintf(stderr, "Cg vertex profile: arbvp1.\n");
  // try VP30
  } else if (cgGLIsProfileSupported(CG_PROFILE_VP30)) {
    vertexProfile = CG_PROFILE_VP30;
    fprintf(stderr, "Cg vertex profile: vp30.\n");
  // try VP20
  } else if (cgGLIsProfileSupported(CG_PROFILE_VP20)) {
    vertexProfile = CG_PROFILE_VP20;
    fprintf(stderr, "Cg vertex profile: vp20.\n");
  } else {
    fprintf(stderr,"Neither arbvp1, vp30, or vp20 vertex profiles supported");
    fprintf(stderr," on this system.\n");
//  exit(1);
  }

  if (cgGLIsProfileSupported(CG_PROFILE_ARBFP1)) {
    fragmentProfile = CG_PROFILE_ARBFP1;
    fprintf(stderr, "Cg fragment profile: arbfp1.\n");
  // try FP30
  } else if (cgGLIsProfileSupported(CG_PROFILE_FP30)) {
    fragmentProfile = CG_PROFILE_FP30;
    fprintf(stderr, "Cg fragment profile: fp30.\n");
  // try FP20
  } else if (cgGLIsProfileSupported(CG_PROFILE_FP20)) {
    fragmentProfile = CG_PROFILE_FP20;
    fprintf(stderr, "Cg fragment profile: fp20.\n");
  } else {
    fprintf(stderr,"Neither arbfp1, fp30, or fp20 fragment profiles supported");
    fprintf(stderr," on this system.\n");
//  exit(1);
  }
}

void loadCgPrograms()
{
  assert(cgIsContext(context));

  // load and compile the vertex program from vert.cg; hold on to the
  // handle to it that is returned
  fprintf(stderr, "Cg vertex program creating on this system.\n");
  vertexProgram = cgCreateProgramFromFile(context,
                                          CG_SOURCE,
                                          "vert.cg",
                                          vertexProfile, NULL, NULL);
  if(!cgIsProgramCompiled(vertexProgram)) {
    fprintf(stderr, "Cg vertex program compiling on this system.\n");
    cgCompileProgram(vertexProgram);
    fprintf(stderr, "Cg vertex program compiled on this system.\n");
  }

  // enable the appropriate vertex profile and load the vertex program
  cgGLEnableProfile(vertexProfile);
  fprintf(stderr, "Cg vertex program loading on this system.\n");
  cgGLLoadProgram(vertexProgram);

  // and similarly set things up for the fragment program
  fprintf(stderr, "Cg fragment program creating on this system.\n");
  fragmentProgram = cgCreateProgramFromFile(context,
                                            CG_SOURCE,
                                            "frag.cg",
                                            fragmentProfile, NULL, NULL);
  if(!cgIsProgramCompiled(fragmentProgram)) {
    fprintf(stderr, "Cg fragment program compiling on this system.\n");
    cgCompileProgram(fragmentProgram);
    fprintf(stderr, "Cg fragment program compiled on this system.\n");
  }

  cgGLEnableProfile(fragmentProfile);
  fprintf(stderr, "Cg fragment program loading on this system.\n");
  cgGLLoadProgram(fragmentProgram);

  cgGLDisableProfile(vertexProfile);
  cgGLDisableProfile(fragmentProfile);
}

void imageTexture()
{
	GLint	error;
	float	*image = new float[RES*RES*4];
	float	*ip = image;

  srand((unsigned int)0x1337);
  for(int i = 0; i < RES; ++i) {
    for(int j = 0; j < RES; ++j) {
      *ip++ = (float)((float)rand() / (float)RAND_MAX);
      *ip++ = (float)((float)rand() / (float)RAND_MAX);
      *ip++ = (float)((float)rand() / (float)RAND_MAX);
      *ip++ = (float)((float)rand() / (float)RAND_MAX);
    }
  }
  glBindTexture(GL_TEXTURE_2D, texNames[0]);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  error = gluBuild2DMipmaps(GL_TEXTURE_2D,
                            GL_RGBA,		// internalFormat
                            RES,RES,		// width,height
                            GL_RGBA,		// format
                            GL_FLOAT,		// type
                            image);		// texels
  if(error)	fprintf(stderr,"%s\n",gluErrorString(error));
  else		fprintf(stderr,"gluBuild2DMipmaps succeeded.\n");
  // tell Cg which texture id should be associated with the sampler2D
  // parameter to the fragment shader
  cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram, "img_tex"),
                          texNames[0]);
  delete [] image;
}

void maskTexture()
{
	float	g,sigma=100.0;
	float	*mask = new float[RES*RES*4];
	float	*ip = mask;

  for(int i = 0; i < RES; ++i) {
    for(int j = 0; j < RES; ++j) {
      g = gauss(i-RES/2,j-RES/2,sigma);
      *ip++ = g;
      *ip++ = g;
      *ip++ = g;
      *ip++ = g;
    }
  }
  glBindTexture(GL_TEXTURE_2D, texNames[1]);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
  glTexImage2D(GL_TEXTURE_2D,0,GL_RGBA,RES,RES,0,GL_RGBA,GL_FLOAT,mask);
  // tell Cg which texture id should be associated with the sampler2D
  // parameter to the fragment shader
  cgGLSetTextureParameter(cgGetNamedParameter(fragmentProgram, "deg_tex"),
                          texNames[1]);
  delete [] mask;
}

float gauss(int i, int j, float sigma)
{
  return(exp((i*i + j*j)/(-2.0*sigma*sigma)));
}

float width()
{
  return(glutGet(GLUT_WINDOW_WIDTH));
}

float height()
{
  return(glutGet(GLUT_WINDOW_HEIGHT));
}
